import math
import random
from typing import Any, Callable, List, NoReturn, Tuple, Union

import numpy as np


# From https://github.com/riflosnake/HumanCursor/blob/main/humancursor/utilities/human_curve_generator.py
# Edited by Vinyzu for more realistic mouse movements
class HumanizeMouseTrajectory:
    def __init__(self, from_point: Tuple[int, int], to_point: Tuple[int, int]) -> None:
        self.from_point = from_point
        self.to_point = to_point

        if self.from_point == self.to_point:
            self.to_point = (self.from_point[0] + 10, self.from_point[1] + 10)

        self.points = self.generate_curve()
        self.points.append(to_point)

    def easeOutQuad(self, n: float) -> float:
        if not 0.0 <= n <= 1.0:
            raise ValueError("Argument must be between 0.0 and 1.0.")
        return -n * (n - 2)

    def generate_curve(self) -> List[Tuple[float, float]]:
        """Generates the curve based on arguments below, default values below are automatically modified to cause randomness"""
        left_boundary = min(self.from_point[0], self.to_point[0]) - 40
        right_boundary = max(self.from_point[0], self.to_point[0]) + 40
        down_boundary = min(self.from_point[1], self.to_point[1]) - 40
        up_boundary = max(self.from_point[1], self.to_point[1]) + 40

        points_distance = math.sqrt((self.from_point[0] - self.to_point[0]) ** 2 + (self.from_point[1] - self.to_point[1]) ** 2)
        internalKnots = self.generate_internal_knots(left_boundary, right_boundary, down_boundary, up_boundary, int(points_distance**0.25))
        points = self.generate_points(internalKnots)
        points = self.distort_points(points, 2, 2, 0.6)

        target_points = int(points_distance // 4) if int(points_distance // 4) > 2 else 2
        points = self.tween_points(points, target_points)
        return points

    def generate_internal_knots(
        self, l_boundary: Union[int, float], r_boundary: Union[int, float], d_boundary: Union[int, float], u_boundary: Union[int, float], knots_count: int
    ) -> Union[List[Tuple[int, int]], NoReturn]:
        """Generates the internal knots of the curve randomly"""
        if not (self.check_if_numeric(l_boundary) and self.check_if_numeric(r_boundary) and self.check_if_numeric(d_boundary) and self.check_if_numeric(u_boundary)):
            raise ValueError("Boundaries must be numeric values")
        if not isinstance(knots_count, int) or knots_count < 0:
            knots_count = 0
        if l_boundary > r_boundary:
            raise ValueError("left_boundary must be less than or equal to right_boundary")
        if d_boundary > u_boundary:
            raise ValueError("down_boundary must be less than or equal to upper_boundary")

        # knotsX = np.random.choice(range(int(l_boundary), int(r_boundary)), size=knots_count)
        # knotsY = np.random.choice(range(int(d_boundary), int(u_boundary)), size=knots_count)
        # knots = list(zip(knotsX, knotsY))

        # Standard deviation for normal distribution
        midpoint_x = (l_boundary + r_boundary) / 2
        midpoint_y = (d_boundary + u_boundary) / 2
        std_deviation = min(midpoint_x - l_boundary, r_boundary - midpoint_x, midpoint_y - d_boundary, u_boundary - midpoint_y)

        # Generate knotsX and knotsY using normal distribution
        knotsX = np.random.normal(midpoint_x, std_deviation, size=knots_count).astype(int)
        knotsY = np.random.normal(midpoint_y, std_deviation, size=knots_count).astype(int)

        # Clip values to within boundaries
        knotsX = np.clip(knotsX, l_boundary, r_boundary)
        knotsY = np.clip(knotsY, d_boundary, u_boundary)
        knots = list(zip(knotsX, knotsY))

        # Sort Knots
        def distance(knot):
            return math.sqrt((self.from_point[0] - knot[0]) ** 2 + (self.from_point[0] - knot[1]) ** 2)

        sorted_knots = sorted(knots, key=lambda knot: distance(knot))
        return sorted_knots

    def generate_points(self, knots: List[Tuple[int, int]]) -> List[Tuple[float, float]]:
        """Generates the points from BezierCalculator"""
        if not self.check_if_list_of_points(knots):
            raise ValueError("knots must be valid list of points")

        midPtsCnt = max(
            abs(self.from_point[0] - self.to_point[0]),
            abs(self.from_point[1] - self.to_point[1]),
            2,
        )
        knots = [self.from_point] + knots + [self.to_point]
        return BezierCalculator.calculate_points_in_curve(int(midPtsCnt), knots)

    def distort_points(self, points: List[Tuple[float, float]], distortion_mean: int, distortion_st_dev: int, distortion_frequency: float) -> Union[List[Tuple[float, float]], NoReturn]:
        """Distorts points by parameters of mean, standard deviation and frequency"""
        if not (self.check_if_numeric(distortion_mean) and self.check_if_numeric(distortion_st_dev) and self.check_if_numeric(distortion_frequency)):
            raise ValueError("Distortions must be numeric")
        if not self.check_if_list_of_points(points):
            raise ValueError("points must be valid list of points")
        if not (0 <= distortion_frequency <= 1):
            raise ValueError("distortion_frequency must be in range [0,1]")

        distorted: List[Tuple[float, float]] = []
        for i in range(1, len(points) - 1):
            x, y = points[i]
            delta = int(np.random.normal(distortion_mean, distortion_st_dev) if random.random() < distortion_frequency else 0)
            distorted.append((x + delta // 5, y + delta // 5))
        distorted = [points[0]] + distorted + [points[-1]]
        return distorted

    def tween_points(self, points: List[Tuple[float, float]], target_points: int) -> Union[List[Tuple[float, float]], NoReturn]:
        """Modifies points by tween"""
        if not self.check_if_list_of_points(points):
            raise ValueError("List of points not valid")
        if not isinstance(target_points, int) or target_points < 2:
            raise ValueError("target_points must be an integer greater or equal to 2")

        res: List[Tuple[float, float]] = []
        for i in range(target_points):
            index = int(self.easeOutQuad(float(i) / (target_points - 1)) * (len(points) - 1))
            res.append(points[index])
        return res

    @staticmethod
    def check_if_numeric(val: Any) -> bool:
        """Checks if value is proper numeric value"""
        return isinstance(val, (float, int, np.integer, np.float32, np.float64))

    def check_if_list_of_points(self, list_of_points: Union[List[Tuple[int, int]], List[Tuple[float, float]]]) -> bool:
        """Checks if list of points is valid"""
        try:

            def point(p):
                return (len(p) == 2) and self.check_if_numeric(p[0]) and self.check_if_numeric(p[1])

            return all(map(point, list_of_points))
        except (KeyError, TypeError):
            return False


class BezierCalculator:
    @staticmethod
    def binomial(n: int, k: int) -> float:
        """Returns the binomial coefficient "n choose k" """
        return math.factorial(n) / float(math.factorial(k) * math.factorial(n - k))

    @staticmethod
    def bernstein_polynomial_point(x: int, i: int, n: int) -> float:
        """Calculate the i-th component of a bernstein polynomial of degree n"""
        return float(BezierCalculator.binomial(n, i) * (x**i) * ((1 - x) ** (n - i)))  # Quirky for mypy :D

    @staticmethod
    def bernstein_polynomial(points: List[Tuple[int, int]]) -> Callable[[float], Tuple[float, float]]:
        """
        Given list of control points, returns a function, which given a point [0,1] returns
        a point in the Bezier described by these points
        """

        def bernstein(t):
            n = len(points) - 1
            x = y = 0.0
            for i, point in enumerate(points):
                bern = BezierCalculator.bernstein_polynomial_point(t, i, n)
                x += point[0] * bern
                y += point[1] * bern
            return x, y

        return bernstein

    @staticmethod
    def calculate_points_in_curve(n: int, points: List[Tuple[int, int]]) -> List[Tuple[float, float]]:
        """
        Given list of control points, returns n points in the Bézier curve,
        described by these points
        """
        curvePoints: List[Tuple[float, float]] = []
        bernstein_polynomial = BezierCalculator.bernstein_polynomial(points)
        for i in range(n):
            t = i / (n - 1)
            curvePoints += (bernstein_polynomial(t),)
        return curvePoints



import platform

VERSION = "1.0.1"

system_name = platform.system()
if system_name == "Windows":
    is_windows = True
elif system_name == "Linux":
    is_windows = False
else:
    raise SystemError("Unknown system (You´re probably using MacOS, which is currently not supported).")

__all__ = ["VERSION", "is_windows"]



import asyncio
import ctypes
import re
import warnings
from typing import Literal, Union

from pywinauto import application, timings
from pywinauto.application import WindowSpecification
from pywinauto.base_wrapper import ElementNotVisible
from pywinauto.controls.hwndwrapper import HwndWrapper, InvalidWindowHandle

from cdp_patches.input.exceptions import WindowClosedException

timings.Timings.fast()
timings.TimeConfig._timings["sendmessagetimeout_timeout"] = 0
timings.TimeConfig._timings["after_click_wait"] = 0
timings.TimeConfig._timings["after_clickinput_wait"] = 0
timings.TimeConfig._timings["after_sendkeys_key_wait"] = 0
# don't block asyncio
timings.TimeConfig._timings["scroll_step_wait"] = 0.01
timings.TimeConfig._timings["window_find_timeout"] = 0.01
timings.TimeConfig._timings["exists_timeout"] = 0.01

warnings.filterwarnings("ignore", category=UserWarning, message="32-bit application should be automated using 32-bit Python (you use 64-bit Python)")


class WindowsBase:
    browser_window: Union[WindowSpecification, HwndWrapper]
    hwnd: int
    pid: int
    scale_factor: float = 1.0
    toolbar_height: int = 0
    win32_app: application.Application = None

    def __init__(self, pid: int, scale_factor: float) -> None:
        self.pid = pid
        self.scale_factor = scale_factor
        self._loop = asyncio.get_event_loop()

    def include_windows_scale_factor(self):
        windows_scale_factor = ctypes.windll.shcore.GetScaleFactorForDevice(0) / 100
        self.scale_factor *= windows_scale_factor

    def get_window(
        self,
        timeout: float = 1,
    ) -> WindowSpecification:
        if self.win32_app is None:
            # save time
            self.win32_app = application.Application(backend="win32")
            self.win32_app.connect(process=self.pid, timeout=timeout)
        try:
            windows = self.win32_app.windows()
        except InvalidWindowHandle:
            raise WindowClosedException(pid=self.pid)
        if not windows:
            raise WindowClosedException(f"No windows found for PID: {self.pid}")
        for window in windows:
            if window.element_info.class_name == "Chrome_WidgetWin_1" and window.is_visible():
                self.browser_window = window
                break
        else:
            # win32_app.top_window(), but without timeout
            criteria = {"backend": self.win32_app.backend.name}
            if windows[0].handle:
                criteria["handle"] = windows[0].handle
            else:
                criteria["name"] = windows[0].name
            self.browser_window = WindowSpecification(criteria, allow_magic_lookup=self.win32_app.allow_magic_lookup)

        for child in self.browser_window.iter_children():
            if child.element_info.class_name == "Chrome_RenderWidgetHostHWND":
                self.browser_window = child

        self.hwnd = self.browser_window.handle
        # Perform Window Checks
        try:
            self.browser_window.verify_actionable()
        except ElementNotVisible:
            raise WindowClosedException(pid=self.pid)
        if not self.browser_window.is_normal():
            raise WindowClosedException(pid=self.pid)

        return self.browser_window

    def ensure_window(self) -> None:
        try:
            if not self.browser_window.is_visible():
                self.get_window()
        except InvalidWindowHandle:
            self.get_window()

    async def async_get_window(self, timeout: float = 1) -> WindowSpecification:
        if self.win32_app is None:
            # save time
            self.win32_app = application.Application(backend="win32")
            await self._loop.run_in_executor(None, lambda: self.win32_app.connect(process=self.pid, timeout=timeout))

        windows = self.win32_app.windows()
        if not windows:
            raise WindowClosedException(f"No windows found for PID: {self.pid}")
        for window in windows:
            if window.element_info.class_name == "Chrome_WidgetWin_1" and window.is_visible():
                self.browser_window = window
                break
        else:
            # top window, but without timeout
            criteria = {"backend": self.win32_app.backend.name}
            if windows[0].handle:
                criteria["handle"] = windows[0].handle
            else:
                criteria["name"] = windows[0].name
            self.browser_window = WindowSpecification(criteria, allow_magic_lookup=self.win32_app.allow_magic_lookup)

        for child in self.browser_window.iter_children():
            if child.element_info.class_name == "Chrome_RenderWidgetHostHWND":
                self.browser_window = child

        self.hwnd = self.browser_window.handle
        # Perform Window Checks
        try:
            self.browser_window.verify_actionable()
        except ElementNotVisible:
            raise WindowClosedException(pid=self.pid)
        if not self.browser_window.is_normal():
            raise WindowClosedException(pid=self.pid)

        return self.browser_window

    def down(self, button: Literal["left", "right", "middle"], x: int, y: int) -> None:
        self.ensure_window()
        self.browser_window.press_mouse(button=button, coords=(int(x * self.scale_factor), int(y * self.scale_factor)))

    def up(self, button: Literal["left", "right", "middle"], x: int, y: int) -> None:
        self.ensure_window()
        self.browser_window.release_mouse(button=button, coords=(int(x * self.scale_factor), int(y * self.scale_factor)))

    def move(self, x: int, y: int) -> None:
        self.ensure_window()
        self.browser_window.move_mouse(coords=(int(x * self.scale_factor), int(y * self.scale_factor)), pressed="left")

    def scroll(self, direction: Literal["up", "down", "left", "right"], amount: int) -> None:
        self.ensure_window()
        self.browser_window.scroll(direction=direction, amount="line", count=int(amount * self.scale_factor))

    def send_keystrokes(self, text: str) -> None:
        # Unmodifying Pywinauto's modifiers & adding down/up modifiers
        modified_text = ""
        selective_regex = re.compile(r"{[^{}]*}|.")

        for key in selective_regex.findall(text):
            if key in ["+", "^", "%", "~"]:
                key = "{" + key + "}"

            modified_text += key

        self.ensure_window()
        self.browser_window.send_keystrokes(modified_text)



import asyncio

import pytest
from selenium_driverless.types.by import By
from selenium_driverless.types.webelement import WebElement
from selenium_driverless.webdriver import Chrome

from cdp_patches.input.exceptions import WindowClosedException
from tests.server import Server

# from input import KeyboardCodes


async def get_locator_pos(locator: WebElement):
    location = await locator.location
    size = await locator.size
    assert location, size

    x, y, width, height = location.get("x"), location.get("y"), size.get("width"), size.get("height")
    assert x and y and width and height

    x, y = x + width // 2, y + height // 2
    return x, y


@pytest.mark.asyncio
async def test_input_leak(async_driver: Chrome, server: Server) -> None:
    await async_driver.get(server.PREFIX + "/input/button.html")
    await async_driver.execute_script(
        """
        const click_elem = document.querySelector("button")
        window.is_leaking = new Promise((resolve, reject) => {callback = resolve});
        const on_click = async function(e){
                var is_bot = (e.pageY == e.screenY && e.pageX == e.screenX)
                if (is_bot && 1 >= outerHeight - innerHeight){ // fullscreen
                    is_bot = false
                }
                callback(is_bot)
        }
        click_elem.removeEventListener("mousedown", self)
        click_elem.addEventListener("mousedown", on_click)
    """
    )
    sync_locator = await async_driver.find_element(By.XPATH, "//button")
    x, y = await get_locator_pos(sync_locator)
    await async_driver.async_input.click("left", x, y)  # type: ignore[attr-defined]

    is_leaking = await async_driver.eval_async("return await window.is_leaking", timeout=300)
    assert not is_leaking


@pytest.mark.asyncio
async def test_click_the_button(async_driver: Chrome, server: Server) -> None:
    await async_driver.get(server.PREFIX + "/input/button.html")
    sync_locator = await async_driver.find_element(By.XPATH, "//button")
    x, y = await get_locator_pos(sync_locator)
    await async_driver.async_input.click("left", x, y)  # type: ignore[attr-defined]
    assert await async_driver.execute_script("return result") == "Clicked"


@pytest.mark.asyncio
async def test_double_click_the_button(async_driver: Chrome, server: Server) -> None:
    await async_driver.get(server.PREFIX + "/input/button.html")
    await async_driver.execute_script(
        """window.double = false;
            button = document.querySelector('button');
            button.addEventListener('dblclick', event => window.double = true);"""
    )

    sync_locator = await async_driver.find_element(By.XPATH, "//button")
    x, y = await get_locator_pos(sync_locator)
    await async_driver.async_input.double_click("left", x, y)  # type: ignore[attr-defined]
    assert await async_driver.execute_script("return window.double")
    assert await async_driver.execute_script("return result") == "Clicked"


@pytest.mark.asyncio
async def test_locators_hover(async_driver: Chrome, server: Server) -> None:
    # x = """window.addEventListener("DOMContentLoaded",()=>{function highest_z_idx(){const allElements=document.querySelectorAll("*");let highestZIndex=0;allElements.forEach((element)=>{const
    #     zIndex=parseInt(getComputedStyle(element).zIndex,10);if(zIndex&&zIndex>highestZIndex){highestZIndex=zIndex}});return highestZIndex}function round_2(num){return Math.round((
    #     num+Number.EPSILON)*100)/100}z_idx=highest_z_idx();const canvas=document.createElement("canvas");canvas.style.position="fixed";canvas.style.top="0";canvas.style.left="0";canvas.style.zIndex=
    #     String(z_idx+1);canvas.width=window.innerWidth;canvas.height=window.innerHeight;canvas.style.pointerEvents="none";const clearButton=document.createElement("button");clearButton.textContent=
    #     "Clear";clearButton.style.position="fixed";clearButton.style.top="10px";clearButton.style.left="10px";clearButton.id="clear";clearButton.style.zIndex=String(z_idx+2);clearButton.style.opacity=
    #     "0.7";const tab=document.createElement("div");tab.style.position="fixed";tab.style.top="10px";tab.style.right="10px";tab.style.padding="5px 10px";tab.style.borderRadius="5px";
    #     tab.style.pointerEvents="none";tab.style.fontFamily="Arial, sans-serif";tab.style.fontSize="14px";tab.style.fontWeight="bold";tab.style.zIndex=String(z_idx+3);tab.style.opacity="0.8";
    #     tab.textContent="average ClickDeltaTime: 0.00ms +/-0.00,Average Frequency: 0.00 Hz, count:0, x:0, y:0";const graphCanvas=document.createElement("canvas");graphCanvas.width=window.innerWidth;
    #     graphCanvas.height=200;graphCanvas.style.position="fixed";graphCanvas.style.bottom="0";graphCanvas.style.left="0";graphCanvas.style.zIndex=String(z_idx+4);graphCanvas.style.pointerEvents=
    #     "None";const ctx=canvas.getContext("2d");let lastEventTime=0;let timeSinceClear=0;let timeDeltaData=[];let mousedownTime=0;let clickdeltaTimes=[];let averageClickDeltaTime=0;
    #     click_delta_max_diff=0;function plot_point(x,y,color="red",radius="2",opacity=0.5){ctx.fillStyle=color;ctx.globalAlpha=opacity;ctx.beginPath();ctx.arc(x,y,radius,0,Math.PI*2);ctx.fill();
    #     ctx.globalAlpha=1;}function clear(){ctx.clearRect(0,0,canvas.width,canvas.height);lastEventTime=0;timeSinceClear=0;timeDeltaData=[];let mousedownTime=0;let clickdeltaTimes=[];
    #     let averageClickDeltaTime=0;click_delta_max_diff=0;tab.textContent="average ClickDeltaTime: 0.00ms +/-0.00, Average Frequency: 0.00 Hz, count:0, x:0, y:0"};function updateCanvasDimensions()
    #     {canvas.width=window.innerWidth;canvas.height=window.innerHeight;graphCanvas.width=window.innerWidth;drawTimeDeltaGraph()}function drawTimeDeltaGraph(){const graphCtx=
    #     graphCanvas.getContext("2d");graphCtx.clearRect(0,0,graphCanvas.width,graphCanvas.height);graphCtx.globalAlpha=0.3;graphCtx.fillStyle="white";graphCtx.fillRect(0,0,graphCanvas.width,
    #     graphCanvas.height);graphCtx.globalAlpha=1;graphCtx.fillStyle="black";if(timeDeltaData.length>graphCanvas.width){timeDeltaData.splice(0,timeDeltaData.length-graphCanvas.width)}
    #     const maxTimeDelta=Math.max(...timeDeltaData);const scaleFactor=graphCanvas.height/maxTimeDelta;const gridSpacing=20;graphCtx.strokeStyle="black";graphCtx.beginPath();for(let y=0;y<=
    #     graphCanvas.height;y+=gridSpacing){graphCtx.moveTo(0,y);graphCtx.lineTo(graphCanvas.width,y);const timeValue=(maxTimeDelta*(graphCanvas.height-y)/graphCanvas.height).toFixed(2);
    #     if(isFinite(timeValue)){graphCtx.fillText(timeValue+" ms",graphCanvas.width-50,y+12)}}graphCtx.stroke();graphCtx.beginPath();graphCtx.strokeStyle="black";graphCtx.moveTo(0,0);
    #     graphCtx.lineTo(graphCanvas.width,0);graphCtx.stroke();graphCtx.beginPath();graphCtx.strokeStyle="black";graphCtx.moveTo(0,graphCanvas.height);graphCtx.lineTo(graphCanvas.width,
    #     graphCanvas.height);graphCtx.stroke();graphCtx.beginPath();graphCtx.strokeStyle="green";graphCtx.moveTo(0,graphCanvas.height-timeDeltaData[0]*scaleFactor);for(let i=1;i<
    #     timeDeltaData.length;i+=1){graphCtx.lineTo(i,graphCanvas.height-timeDeltaData[i]*scaleFactor)}graphCtx.stroke();graphCtx.fillStyle="black";graphCtx.font="12px Arial";
    #     graphCtx.fillText("0 ms",2,graphCanvas.height-2);graphCtx.fillText(`${timeDeltaData.length-1} ms`,graphCanvas.width-30,graphCanvas.height-2);graphCtx.fillText(`${maxTimeDelta.toFixed(2)} ms`
    #     ,2,10)}function move_handler(event){const currentTime=Date.now();const delta=currentTime-lastEventTime;const x=event.x;const y=event.y;if(delta<=100){if(lastEventTime!==0){timeSinceClear+=
    #     delta;timeDeltaData.push(delta);const averageDelta=timeDeltaData.length===0?0:timeDeltaData.reduce((sum,value)=>sum+value)/timeDeltaData.length;const frequency=averageDelta===0?0:1000/
    #     averageDelta;tab.textContent=`average ClickDeltaTime: ${ averageClickDeltaTime }ms +/-${ click_delta_max_diff }, Average Frequency: ${frequency.toFixed(2)} Hz, count:${timeDeltaData.length }
    #     ,x:${ x }, y:${ y }`;drawTimeDeltaGraph()}}lastEventTime=currentTime;plot_point(x,y)}function click_handler(e){plot_point(e.x,e.y,"green",5)};function mouseup_handler(){const mouseupTime=
    #     Date.now();const deltaTime=mouseupTime-mousedownTime;clickdeltaTimes.push(deltaTime);delta_average=clickdeltaTimes.reduce((sum,time)=>sum+time,
    #     0)/clickdeltaTimes.length;click_delta_max_diff=round_2(((Math.max(...clickdeltaTimes)-delta_average)+(delta_average-Math.min(...clickdeltaTimes)))/2);averageClickDeltaTime=
    #     round_2(delta_average)}document.body.appendChild(canvas);document.body.appendChild(graphCanvas);document.addEventListener("mousemove",move_handler);document.addEventListener("click",
    #     click_handler);document.addEventListener("mousedown",(e)=>{mousedownTime=Date.now()});document.body.appendChild(tab);window.addEventListener("resize",updateCanvasDimensions);
    #     document.body.appendChild(clearButton);clearButton.addEventListener("click",clear);updateCanvasDimensions();document.addEventListener("mouseup",mouseup_handler);})"""
    # sync_page.add_init_script(x)

    await async_driver.get(server.PREFIX + "/input/scrollable.html")
    await async_driver.async_input.move(500, 100)  # type: ignore[attr-defined]

    sync_locator = await async_driver.find_element(By.ID, "button-12")
    x, y = await get_locator_pos(sync_locator)
    await async_driver.async_input.move(x, y)  # type: ignore[attr-defined]

    await asyncio.sleep(0.5)
    assert await async_driver.execute_script("return document.querySelector('button:hover').id") == "button-12"


@pytest.mark.asyncio
async def test_fill_input(async_driver: Chrome, server: Server) -> None:
    await async_driver.get(server.PREFIX + "/input/textarea.html")
    sync_locator = await async_driver.find_element(By.XPATH, "//input")
    assert sync_locator

    x, y = await get_locator_pos(sync_locator)
    await async_driver.async_input.click("left", x, y)  # type: ignore[attr-defined]
    await async_driver.async_input.type("some value", fill=True)  # type: ignore[attr-defined]
    assert await async_driver.execute_script("return result") == "some value"


@pytest.mark.asyncio
async def test_keyboard_type_into_a_textarea(async_driver: Chrome) -> None:
    await async_driver.execute_script(
        """
            const textarea = document.createElement('textarea');
            document.body.appendChild(textarea);
            textarea.focus();
        """
    )
    text = "Hello world. I +am  the %text that was typed!"

    sync_locator = await async_driver.find_element(By.XPATH, "//textarea")
    assert sync_locator

    x, y = await get_locator_pos(sync_locator)
    await async_driver.async_input.click("left", x, y)  # type: ignore[attr-defined]

    await async_driver.async_input.type(text)  # type: ignore[attr-defined]
    assert await async_driver.execute_script('return document.querySelector("textarea").value') == text


@pytest.mark.asyncio
async def test_quit_exception(async_driver: Chrome) -> None:
    await async_driver.quit()
    await asyncio.sleep(5)

    with pytest.raises(WindowClosedException):
        await async_driver.async_input.down("left", 100, 100, emulate_behaviour=False)
    with pytest.raises(WindowClosedException):
        await async_driver.async_input.up("left", 110, 110)
    with pytest.raises(WindowClosedException):
        await async_driver.async_input.move(50, 50, emulate_behaviour=False)
    with pytest.raises(WindowClosedException):
        await async_driver.async_input.scroll("up", 10)
    with pytest.raises(WindowClosedException):
        await async_driver.async_input.type("test")



import json
import time
from contextlib import suppress
from typing import Dict, List, Type, TypedDict, Union

import requests
from websockets.sync import client

try:
    from playwright.async_api import Browser as AsyncBrowser
    from playwright.async_api import BrowserContext as AsyncContext
    from playwright.async_api import Error as AsyncError
    from playwright.async_api import Error as SyncError
    from playwright.sync_api import Browser as SyncBrowser
    from playwright.sync_api import BrowserContext as SyncContext
except ImportError:
    AsyncBrowser: Type["AsyncBrowser"] = "AsyncBrowser"  # type: ignore[no-redef]
    AsyncContext: Type["AsyncContext"] = "AsyncContext"  # type: ignore[no-redef]
    SyncBrowser: Type["SyncBrowser"] = "SyncBrowser"  # type: ignore[no-redef]
    SyncContext: Type["SyncContext"] = "SyncContext"  # type: ignore[no-redef]

try:
    from botright.extended_typing import BrowserContext as BotrightContext
except ImportError:
    BotrightContext: Type["BotrightContext"] = "BotrightContext"  # type: ignore[no-redef]

try:
    from selenium.webdriver import Chrome as SeleniumChrome
except ImportError:
    SeleniumChrome: Type["SeleniumChrome"] = "SeleniumChrome"  # type: ignore[no-redef]

try:
    from selenium_driverless.sync.webdriver import Chrome as DriverlessSyncChrome
    from selenium_driverless.webdriver import Chrome as DriverlessAsyncChrome
except ImportError:
    DriverlessAsyncChrome: Type["DriverlessAsyncChrome"] = "DriverlessAsyncChrome"  # type: ignore[no-redef]
    DriverlessSyncChrome: Type["DriverlessSyncChrome"] = "DriverlessSyncChrome"  # type: ignore[no-redef]

all_browsers = Union[AsyncContext, AsyncBrowser, SyncContext, SyncBrowser, BotrightContext, SeleniumChrome, DriverlessAsyncChrome, DriverlessSyncChrome]
sync_browsers = Union[SeleniumChrome, SyncContext, SyncBrowser, DriverlessSyncChrome]
async_browsers = Union[AsyncContext, AsyncBrowser, BotrightContext, DriverlessAsyncChrome]


class InternalProcessInfo(TypedDict):
    type: str
    id: int
    cpuTime: float


class CDPProcessInfo:
    processInfo: List[InternalProcessInfo]

    def __init__(self, process_info: Dict[str, List[InternalProcessInfo]]) -> None:
        self.processInfo = process_info["processInfo"]

    def get_main_browser(self) -> InternalProcessInfo:
        for process in self.processInfo:
            if process.get("type") == "browser":
                return process

        raise ValueError("No browser process found.")


class SeleniumProcessInfoResponse(TypedDict):
    id: int
    result: Dict[str, List[InternalProcessInfo]]


def process_info_from_ws(ws_url: str) -> Dict[str, List[InternalProcessInfo]]:
    with client.connect(ws_url) as websocket:
        websocket.send(json.dumps({"id": 1, "method": "SystemInfo.getProcessInfo"}))
        for message in websocket:
            data: SeleniumProcessInfoResponse = json.loads(message)
            if data.get("id") == 1:
                return data["result"]

    raise RuntimeError("Couldn't connect to browser.")


def ws_url_from_url(url: str, timeout: float = 30) -> str:
    if len(url) < 7 or url[:7] != "http://":
        url = "http://" + url + "/json/version"
    try:
        data = requests.get(url, timeout=timeout).json()
    except requests.exceptions.Timeout:
        raise TimeoutError(f"Couldn't connect to browser within {timeout} seconds")

    websocket_debugger_url: str = data["webSocketDebuggerUrl"]
    return websocket_debugger_url


def process_info_from_url(url: str) -> Dict[str, List[InternalProcessInfo]]:
    ws_url = ws_url_from_url(url)
    return process_info_from_ws(ws_url)


# Browser PID
# Selenium & Selenium Driverless
def get_sync_selenium_browser_pid(driver: Union[SeleniumChrome, DriverlessSyncChrome]) -> int:
    if isinstance(driver, DriverlessSyncChrome):
        cdp_system_info = driver.base_target.execute_cdp_cmd(cmd="SystemInfo.getProcessInfo")
    elif isinstance(driver, SeleniumChrome):
        cdp_system_info = process_info_from_url(driver.capabilities["goog:chromeOptions"]["debuggerAddress"])
    else:
        raise ValueError("Invalid browser type.")
    process_info = CDPProcessInfo(cdp_system_info)
    browser_info = process_info.get_main_browser()
    return browser_info["id"]


async def get_async_selenium_browser_pid(driver: DriverlessAsyncChrome) -> int:
    cdp_system_info = await driver.base_target.execute_cdp_cmd(cmd="SystemInfo.getProcessInfo")

    process_info = CDPProcessInfo(cdp_system_info)
    browser_info = process_info.get_main_browser()
    return browser_info["id"]


# Playwright
def get_sync_playwright_browser_pid(browser: Union[SyncContext, SyncBrowser]) -> int:
    if isinstance(browser, SyncContext):
        main_browser = browser.browser
        assert main_browser
        cdp_session = main_browser.new_browser_cdp_session()
    elif isinstance(browser, SyncBrowser):
        cdp_session = browser.new_browser_cdp_session()
    else:
        raise ValueError("Invalid browser type.")

    cdp_system_info = cdp_session.send("SystemInfo.getProcessInfo")

    process_info = CDPProcessInfo(cdp_system_info)
    browser_info = process_info.get_main_browser()
    return browser_info["id"]


async def get_async_playwright_browser_pid(browser: Union[AsyncContext, AsyncBrowser, BotrightContext]) -> int:
    if isinstance(browser, AsyncContext) or isinstance(browser, BotrightContext):
        main_browser = browser.browser
        assert main_browser
        cdp_session = await main_browser.new_browser_cdp_session()
    elif isinstance(browser, AsyncBrowser):
        cdp_session = await browser.new_browser_cdp_session()
    else:
        raise ValueError("Invalid browser type.")
    cdp_system_info = await cdp_session.send("SystemInfo.getProcessInfo")

    process_info = CDPProcessInfo(cdp_system_info)
    browser_info = process_info.get_main_browser()
    return browser_info["id"]


def get_sync_browser_pid(browser: sync_browsers) -> int:
    if isinstance(browser, SeleniumChrome) or isinstance(browser, DriverlessSyncChrome):
        return get_sync_selenium_browser_pid(browser)
    elif isinstance(browser, SyncContext) or isinstance(browser, SyncBrowser):
        return get_sync_playwright_browser_pid(browser)

    raise ValueError("Invalid browser type.")


async def get_async_browser_pid(browser: async_browsers) -> int:
    if isinstance(browser, DriverlessAsyncChrome):
        return await get_async_selenium_browser_pid(browser)
    elif isinstance(browser, AsyncContext) or isinstance(browser, AsyncBrowser) or isinstance(browser, BotrightContext):
        return await get_async_playwright_browser_pid(browser)

    raise ValueError("Invalid browser type.")


# Scale Factor
# Selenium & Selenium Driverless
def get_sync_selenium_scale_factor(driver: Union[SeleniumChrome, DriverlessSyncChrome]) -> int:
    if isinstance(driver, DriverlessSyncChrome):
        _scale_factor: int = driver.execute_script("return window.devicePixelRatio", unique_context=True)
        return _scale_factor

    scale_factor: int = driver.execute_script("return window.devicePixelRatio")
    return scale_factor


async def get_async_selenium_scale_factor(driver: DriverlessAsyncChrome) -> int:
    scale_factor: int = await driver.execute_script("return window.devicePixelRatio", unique_context=True)
    return scale_factor


# Playwright with Runtime Patching
def get_sync_playwright_scale_factor(browser: Union[SyncContext, SyncBrowser]) -> int:
    close_context, close_page = False, False
    if isinstance(browser, SyncContext):
        context = browser
    elif isinstance(browser, SyncBrowser):
        if any(browser.contexts):
            context = browser.contexts[0]
        else:
            context = browser.new_context()
            close_context = True
    else:
        raise ValueError("Invalid browser type.")

    if any(context.pages):
        page = context.pages[0]
    else:
        page = context.new_page()
        close_page = True
    cdp_session = context.new_cdp_session(page)

    time1 = time.perf_counter()
    while (time.perf_counter() - time1) <= 10:
        try:
            page_frame_tree = cdp_session.send("Page.getFrameTree")
            page_id = page_frame_tree["frameTree"]["frame"]["id"]

            isolated_world = cdp_session.send("Page.createIsolatedWorld", {"frameId": page_id, "grantUniveralAccess": True, "worldName": "Shimmy shimmy yay, shimmy yay, shimmy ya"})
            isolated_exec_id = isolated_world["executionContextId"]
            break
        except SyncError as e:
            if e.message == "Protocol error (Page.createIsolatedWorld): Invalid parameters":
                pass
            else:
                raise e
    else:
        raise TimeoutError("Page.createIsolatedWorld did not initialize properly within 30 seconds.")

    time2 = time.perf_counter()
    while (time.perf_counter() - time2) <= 10:
        try:
            scale_factor_eval = cdp_session.send("Runtime.evaluate", {"expression": "window.devicePixelRatio", "contextId": isolated_exec_id})
            scale_factor: int = scale_factor_eval["result"]["value"]
            break
        except SyncError as e:
            if e.message == "Protocol error (Runtime.evaluate): Cannot find context with specified id":
                pass
            else:
                raise e
    else:
        raise TimeoutError("Runtime.evaluate did not run properly within 30 seconds.")

    with suppress(SyncError):
        if close_page:
            page.close()

    with suppress(SyncError):
        if close_context:
            context.close()

    return scale_factor


async def get_async_playwright_scale_factor(browser: Union[AsyncContext, AsyncBrowser, BotrightContext]) -> int:
    close_context, close_page = False, False
    if isinstance(browser, AsyncContext) or isinstance(browser, BotrightContext):
        context = browser
    elif isinstance(browser, AsyncBrowser):
        if any(browser.contexts):
            context = browser.contexts[0]
        else:
            context = await browser.new_context()
            close_context = True
    else:
        raise ValueError("Invalid browser type.")

    if any(context.pages):
        page = context.pages[0]
    else:
        page = await context.new_page()
        close_page = True
    cdp_session = await context.new_cdp_session(page)

    time1 = time.perf_counter()
    while (time.perf_counter() - time1) <= 10:
        try:
            page_frame_tree = await cdp_session.send("Page.getFrameTree")
            page_id = page_frame_tree["frameTree"]["frame"]["id"]

            isolated_world = await cdp_session.send("Page.createIsolatedWorld", {"frameId": page_id, "grantUniveralAccess": True, "worldName": "Shimmy shimmy yay, shimmy yay, shimmy ya"})
            isolated_exec_id = isolated_world["executionContextId"]
            break
        except AsyncError as e:
            if e.message == "Protocol error (Page.createIsolatedWorld): Invalid parameters":
                pass
            else:
                raise e
    else:
        raise TimeoutError("Page.createIsolatedWorld did not initialize properly within 30 seconds.")

    time2 = time.perf_counter()
    while (time.perf_counter() - time2) <= 10:
        try:
            scale_factor_eval = await cdp_session.send("Runtime.evaluate", {"expression": "window.devicePixelRatio", "contextId": isolated_exec_id})
            scale_factor: int = scale_factor_eval["result"]["value"]
            break
        except AsyncError as e:
            if e.message == "Protocol error (Runtime.evaluate): Cannot find context with specified id":
                pass
            else:
                raise e
    else:
        raise TimeoutError("Runtime.evaluate did not run properly within 30 seconds.")

    with suppress(SyncError):
        if close_page:
            await page.close()

    with suppress(SyncError):
        if close_context:
            await context.close()

    return scale_factor


def get_sync_scale_factor(browser: sync_browsers) -> int:
    if isinstance(browser, SeleniumChrome) or isinstance(browser, DriverlessSyncChrome):
        return get_sync_selenium_scale_factor(browser)
    elif isinstance(browser, SyncContext) or isinstance(browser, SyncBrowser):
        return get_sync_playwright_scale_factor(browser)

    raise ValueError("Invalid browser type.")


async def get_async_scale_factor(browser: async_browsers) -> int:
    if isinstance(browser, DriverlessAsyncChrome):
        return await get_async_selenium_scale_factor(browser)
    elif isinstance(browser, AsyncContext) or isinstance(browser, AsyncBrowser) or isinstance(browser, BotrightContext):
        return await get_async_playwright_scale_factor(browser)

    raise ValueError("Invalid browser type.")


__all__ = ["SeleniumChrome", "DriverlessSyncChrome", "DriverlessAsyncChrome"]



import random
import re
import sys
import threading
import time
import warnings
from typing import Literal, Optional, Union

if sys.version_info.minor >= 10:
    from typing import TypeAlias
else:
    TypeAlias = "TypeAlias"  # type: ignore[assignment]

from cdp_patches import is_windows
from cdp_patches.input.exceptions import WindowClosedException

if is_windows:
    from pywinauto.application import ProcessNotFoundError
    from pywinauto.base_wrapper import ElementNotEnabled

    from cdp_patches.input.os_base.windows import WindowsBase  # type: ignore[assignment]

    LinuxBase: TypeAlias = WindowsBase  # type: ignore[no-redef]
    InputBase = WindowsBase  # type: ignore
    WindowErrors = (ValueError, ElementNotEnabled, ProcessNotFoundError, WindowClosedException)  # type: ignore[assignment]
else:
    from cdp_patches.input.os_base.linux import LinuxBase  # type: ignore[assignment]

    WindowsBase: TypeAlias = LinuxBase  # type: ignore[no-redef]
    InputBase = LinuxBase  # type: ignore
    WindowErrors = (AssertionError, ValueError, WindowClosedException)  # type: ignore[assignment]

from .browsers import DriverlessSyncChrome, SeleniumChrome, get_sync_browser_pid, get_sync_scale_factor, sync_browsers
from .mouse_trajectory import HumanizeMouseTrajectory


class SyncInput:
    emulate_behaviour: Optional[bool] = True
    pid: Optional[int]
    # _base: Union[WindowsBase, LinuxBase]
    _base: InputBase
    window_timeout: float = 30
    _scale_factor: float = 1.0
    sleep_timeout: float = 0.01
    typing_speed: int = 50
    last_x: int = 0
    last_y: int = 0
    selective_modifiers_regex = re.compile(r"{[^{}]*}|.")

    def __init__(
        self, pid: Optional[int] = None, browser: Optional[sync_browsers] = None, scale_factor: Optional[float] = 1.0, emulate_behaviour: Optional[bool] = True, window_timeout: Optional[float] = 30.0
    ) -> None:
        self._scale_factor = scale_factor or self._scale_factor
        self.window_timeout = window_timeout or self.window_timeout
        self.emulate_behaviour = emulate_behaviour or self.emulate_behaviour
        self._move_lock = threading.Lock()

        if browser:
            self.pid = get_sync_browser_pid(browser)
            self._scale_factor = float(get_sync_scale_factor(browser))
        elif pid:
            self.pid = pid
        else:
            raise ValueError("You must provide a pid or a browser")

        self._base = InputBase(self.pid, self._scale_factor)  # type: ignore
        self._wait_for_window()

        # Include Windows Scale Factor for every browser except DriverlessSyncChrome
        if is_windows and not isinstance(browser, (DriverlessSyncChrome, SeleniumChrome)):
            self._base.include_windows_scale_factor()

    @property
    def base(self) -> Union[WindowsBase, LinuxBase]:
        return self._base

    @property
    def scale_factor(self) -> float:
        return self._scale_factor

    @scale_factor.setter
    def scale_factor(self, scale_value: float) -> None:
        self._scale_factor = scale_value
        if self._base:
            self._base.scale_factor = scale_value

    def _wait_for_window(self) -> None:
        max_wait = time.perf_counter() + self.window_timeout
        while time.perf_counter() < max_wait:
            try:
                if self._base.get_window():
                    return
            except WindowErrors:
                pass
            self._sleep_timeout(0.1)

        raise TimeoutError(f"Chrome Window (PID: {self.pid}) not found in {self.window_timeout} seconds.")

    def _sleep_timeout(self, timeout: Optional[float] = None) -> None:
        timeout = timeout or self.sleep_timeout
        # if not random.randint(0, 10):
        # timeout_random = timeout / 10
        # timeout = random.uniform(timeout - timeout_random, timeout + timeout_random)

        time.sleep(timeout)
        # Perfect Precise Sleep
        # start = time.perf_counter()
        # while time.perf_counter() - start < timeout:
        #     pass

    def click(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = 0.07) -> None:
        x, y = int(x), int(y)

        self.down(button=button, x=x, y=y, emulate_behaviour=emulate_behaviour, timeout=timeout)
        if self.emulate_behaviour and emulate_behaviour:
            self._sleep_timeout(timeout=timeout)
        self.up(button=button, x=x, y=y)
        self.last_x, self.last_y = x, y

    def double_click(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None) -> None:
        x, y = int(x), int(y)

        self.click(button=button, x=x, y=y, timeout=timeout, emulate_behaviour=emulate_behaviour)
        if emulate_behaviour and self.emulate_behaviour:
            self._sleep_timeout(random.uniform(0.14, 0.21))
            # self._sleep_timeout(timeout=timeout)
        self.click(button=button, x=x, y=y, emulate_behaviour=False, timeout=timeout)

        self.last_x, self.last_y = x, y

    def down(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None) -> None:
        x, y = int(x), int(y)

        if self.emulate_behaviour and emulate_behaviour:
            self.move(x=x, y=y, emulate_behaviour=emulate_behaviour, timeout=timeout)
        self._base.down(button=button, x=x, y=y)
        self.last_x, self.last_y = x, y

    def up(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float]) -> None:
        x, y = int(x), int(y)

        self._base.up(button=button, x=x, y=y)
        self.last_x, self.last_y = x, y

    def move(self, x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None) -> None:
        with self._move_lock:
            x, y = int(x), int(y)

            if self.emulate_behaviour and emulate_behaviour:
                humanized_points = HumanizeMouseTrajectory((self.last_x, self.last_y), (x, y))

                # Move Mouse to new random locations
                for i, (human_x, human_y) in enumerate(humanized_points.points):
                    self._base.move(x=int(human_x), y=int(human_y))
                    self._sleep_timeout(timeout=timeout)

            self._base.move(x=x, y=y)
            self.last_x, self.last_y = x, y

    def scroll(self, direction: Literal["up", "down", "left", "right"], amount: int) -> None:
        warnings.warn("Scrolling using CDP-Patches is discouraged as Scroll Inputs dont leak the CDP Domain.", UserWarning)
        self._base.scroll(direction=direction, amount=amount)

    def type(self, text: str, fill: Optional[bool] = False, timeout: Optional[float] = None) -> None:
        if self.emulate_behaviour and not fill:
            for i, char in enumerate(self.selective_modifiers_regex.findall(text)):
                # If new word is started wait some more time
                if i != 0 and text[i - 1] == " ":
                    self._sleep_timeout(timeout=timeout)

                self._base.send_keystrokes(char)
                self._sleep_timeout((random.random() * 10) / self.typing_speed)
        else:
            self._base.send_keystrokes(text)



import asyncio

import pytest
from playwright.async_api import Locator, Page

from cdp_patches.input.exceptions import WindowClosedException
from tests.server import Server

# from input import KeyboardCodes


async def get_locator_pos(locator: Locator):
    bounding_box = await locator.bounding_box()
    assert bounding_box

    x, y, width, height = bounding_box.get("x"), bounding_box.get("y"), bounding_box.get("width"), bounding_box.get("height")
    assert x and y and width and height

    x, y = x + width // 2, y + height // 2
    return x, y


@pytest.mark.asyncio
async def test_input_leak(async_page: Page, server: Server) -> None:
    await async_page.goto(server.PREFIX + "/input/button.html")
    await async_page.evaluate(
        """
        const click_elem = document.querySelector("button")
        window.is_leaking = new Promise((resolve, reject) => {callback = resolve});
        const on_click = async function(e){
                var is_bot = (e.pageY == e.screenY && e.pageX == e.screenX)
                if (is_bot && 1 >= outerHeight - innerHeight){ // fullscreen
                    is_bot = false
                }
                callback(is_bot)
        }
        click_elem.removeEventListener("mousedown", self)
        click_elem.addEventListener("mousedown", on_click)
    """
    )
    async_locator = async_page.locator("button")
    x, y = await get_locator_pos(async_locator)
    await async_page.async_input.click("left", x, y)  # type: ignore[attr-defined]

    is_leaking = await async_page.evaluate("() => window.is_leaking")
    assert not is_leaking


@pytest.mark.asyncio
async def test_click_the_button(async_page: Page, server: Server) -> None:
    await async_page.goto(server.PREFIX + "/input/button.html")
    locator = async_page.locator("button")
    x, y = await get_locator_pos(locator)
    await async_page.async_input.click("left", x, y)  # type: ignore[attr-defined]
    assert await async_page.evaluate("result") == "Clicked"


@pytest.mark.asyncio
async def test_double_click_the_button(async_page: Page, server: Server) -> None:
    await async_page.goto(server.PREFIX + "/input/button.html")
    await async_page.evaluate(
        """() => {
            window.double = false;
            button = document.querySelector('button');
            button.addEventListener('dblclick', event => window.double = true);
        }"""
    )

    locator = async_page.locator("button")
    x, y = await get_locator_pos(locator)
    await async_page.async_input.double_click("left", x, y)  # type: ignore[attr-defined]
    assert await async_page.evaluate("double")
    assert await async_page.evaluate("result") == "Clicked"


@pytest.mark.asyncio
async def test_locators_hover(async_page: Page, server: Server) -> None:
    await async_page.goto(server.PREFIX + "/input/scrollable.html")
    await async_page.async_input.move(500, 100)  # type: ignore[attr-defined]

    button = async_page.locator("#button-12")
    x, y = await get_locator_pos(button)
    await async_page.async_input.move(x, y)  # type: ignore[attr-defined]

    assert await async_page.evaluate("document.querySelector('button:hover').id") == "button-12"


@pytest.mark.skip(reason="Scrolling is discouraged")
@pytest.mark.asyncio
async def test_scroll(async_page: Page, server: Server) -> None:
    await async_page.goto(server.PREFIX + "/offscreenbuttons.html")
    for i in range(11):
        button = await async_page.query_selector(f"#btn{i}")
        assert button
        before = await button.evaluate(
            """button => {
                return button.getBoundingClientRect().right - window.innerWidth
            }"""
        )

        assert before == 10 * i
        # await button.scroll_into_view_if_needed()
        await async_page.async_input.scroll("right", i)  # type: ignore[attr-defined]

        after = await button.evaluate(
            """button => {
                return button.getBoundingClientRect().right - window.innerWidth
            }"""
        )

        assert after <= 0
        await async_page.evaluate("() => window.scrollTo(0, 0)")


@pytest.mark.asyncio
async def test_fill_input(async_page: Page, server: Server) -> None:
    await async_page.goto(server.PREFIX + "/input/textarea.html")
    handle = async_page.locator("input")
    assert handle

    x, y = await get_locator_pos(handle)
    await async_page.async_input.click("left", x, y)  # type: ignore[attr-defined]
    await async_page.async_input.type("some value", fill=True)  # type: ignore[attr-defined]
    assert await async_page.evaluate("result") == "some value"


@pytest.mark.asyncio
async def test_keyboard_type_into_a_textarea(async_page: Page) -> None:
    await async_page.evaluate(
        """
            const textarea = document.createElement('textarea');
            document.body.appendChild(textarea);
            textarea.focus();
        """
    )
    text = "Hello world. I +am  the %text that was typed!"

    handle = async_page.locator("textarea")
    assert handle

    x, y = await get_locator_pos(handle)
    await async_page.async_input.click("left", x, y)  # type: ignore[attr-defined]

    await async_page.async_input.type(text)  # type: ignore[attr-defined]
    assert await async_page.evaluate('document.querySelector("textarea").value') == text


# @pytest.mark.asyncio
# async def test_should_report_shiftkey(async_page: Page, server: Server) -> None:
#     await async_page.goto(server.PREFIX + "/input/keyboard.html")
#     code_for_key = {KeyboardCodes.VK_SHIFT: ["Shift", "16"], KeyboardCodes.VK_CONTROL: ["Control", "17"]}
#     # , KeyboardCodes.VK_MENU: ["Alt", "18"]
#
#     handle = async_page.locator("textarea")
#     assert handle
#
#     x, y = await get_locator_pos(handle)
#     await async_page.async_input.click("left", x, y)   # type: ignore[attr-defined]
#
#     for modifier_key, js_key in code_for_key.items():
#         await async_page.async_input.press_keys(modifier_key)   # type: ignore[attr-defined]
#         # async_page.async_input._base.browser_window.send_keystrokes("{VK_SHIFT down} bruh")   # type: ignore[attr-defined]
#
#         assert (
#             await async_page.evaluate("() => getResult()")
#             == "Keydown: "
#             + js_key[0]
#             + " "
#             + js_key[0]
#             + "Left "
#             + js_key[1]
#             + " ["
#             + js_key[0]
#             + "]"
#         )
#
#         await async_page.async_input.press_keys("!")   # type: ignore[attr-defined]
#         # Shift+! will generate a keypress
#         if js_key[0] == "Shift":
#             assert (
#                 await async_page.evaluate("() => getResult()")
#                 == "Keydown: ! Digit1 49 ["
#                 + js_key[0]
#                 + "]\nKeypress: ! Digit1 33 33 ["
#                 + js_key[0]
#                 + "]"
#             )
#         else:
#             assert (
#                 await async_page.evaluate("() => getResult()")
#                 == "Keydown: ! Digit1 49 [" + js_key[0] + "]"
#             )
#
#         await async_page.async_input.release_keys("!")   # type: ignore[attr-defined]
#         assert (
#             await async_page.evaluate("() => getResult()")
#             == "Keyup: ! Digit1 49 [" + js_key[0] + "]"
#         )
#         await async_page.async_input.release_keys(modifier_key)   # type: ignore[attr-defined]
#         assert (
#             await async_page.evaluate("() => getResult()")
#             == "Keyup: "
#             + js_key[0]
#             + " "
#             + js_key[0]
#             + "Left "
#             + js_key[1]
#             + " []"
#         )


@pytest.mark.asyncio
async def test_quit_exception(async_page: Page) -> None:
    await async_page.close()
    await asyncio.sleep(5)

    with pytest.raises(WindowClosedException):
        await async_page.async_input.down("left", 100, 100, emulate_behaviour=False)  # type: ignore[attr-defined]
    with pytest.raises(WindowClosedException):
        await async_page.async_input.up("left", 110, 110)  # type: ignore[attr-defined]
    with pytest.raises(WindowClosedException):
        await async_page.async_input.move(50, 50, emulate_behaviour=False)  # type: ignore[attr-defined]
    with pytest.raises(WindowClosedException):
        await async_page.async_input.scroll("up", 10)  # type: ignore[attr-defined]
    with pytest.raises(WindowClosedException):
        await async_page.async_input.type("test")  # type: ignore[attr-defined]



from dataclasses import dataclass

from cdp_patches import is_windows

from .async_input import AsyncInput
from .sync_input import SyncInput


# From: https://pywinauto.readthedocs.io/en/latest/code/pywinauto.keyboard.html
@dataclass
class WinKeyboardCodes:
    SPACE: str = "{SPACE}"
    TAB: str = "{TAB}"
    ENTER: str = "{ENTER}"
    BACKSPACE: str = "{BACKSPACE}"
    DELETE: str = "{DELETE}"
    ESCAPE: str = "{ESC}"
    LEFT_ARROW: str = "{LEFT}"
    RIGHT_ARROW: str = "{RIGHT}"
    UP_ARROW: str = "{UP}"
    DOWN_ARROW: str = "{DOWN}"
    HOME: str = "{HOME}"
    END: str = "{END}"
    PAGE_UP: str = "{PGUP}"
    PAGE_DOWN: str = "{PGDN}"
    PRINT_SCREEN: str = "{PRTSC}"
    SCROLLLOCK: str = "{SCROLLLOCK}"
    CAPSLOCK: str = "{CAPSLOCK}"
    NUMLOCK: str = "{NUMLOCK}"
    F1: str = "{F1}"
    F2: str = "{F2}"
    F3: str = "{F3}"
    F4: str = "{F4}"
    F5: str = "{F5}"
    F6: str = "{F6}"
    F7: str = "{F7}"
    F8: str = "{F8}"
    F9: str = "{F9}"
    F10: str = "{F10}"
    F11: str = "{F11}"
    F12: str = "{F12}"
    F13: str = "{F13}"
    F14: str = "{F14}"
    F15: str = "{F15}"
    F16: str = "{F16}"
    F17: str = "{F17}"
    F18: str = "{F18}"
    F19: str = "{F19}"
    F20: str = "{F20}"
    F21: str = "{F21}"
    F22: str = "{F22}"
    F23: str = "{F23}"
    F24: str = "{F24}"
    INSERT: str = "{INSERT}"
    WIN_LEFT: str = "{LWIN}"
    WIN_RIGHT: str = "{RWIN}"
    CONTROL: str = "{VK_CONTROL}"
    SHIFT: str = "{VK_SHIFT}"
    ALT: str = "{VK_MENU}"
    CANCEL: str = "{VK_CANCEL}"
    CLEAR: str = "{VK_CLEAR}"
    EXECUTE: str = "{VK_EXECUTE}"
    KANA: str = "{VK_KANA}"
    KANJI: str = "{VK_KANJI}"
    LCONTROL: str = "{VK_LCONTROL}"
    LMENU: str = "{VK_LMENU}"
    LSHIFT: str = "{VK_LSHIFT}"
    NEXT: str = "{VK_NEXT}"
    PAUSE: str = "{VK_PAUSE}"
    PRIOR: str = "{VK_PRIOR}"
    RCONTROL: str = "{VK_RCONTROL}"
    RETURN: str = "{VK_RETURN}"
    RSHIFT: str = "{VK_RSHIFT}"
    SELECT: str = "{VK_SELECT}"
    MENU_RIGHT: str = "{RMENU}"
    ADD: str = "{VK_ADD}"
    SUBTRACT: str = "{VK_SUBTRACT}"
    DECIMAL: str = "{VK_DECIMAL}"
    MULTIPLY: str = "{VK_MULTIPLY}"
    DIVIDE: str = "{VK_DIVIDE}"
    SEPARATOR: str = "{VK_SEPARATOR}"
    NUMPAD0: str = "{VK_NUMPAD0}"
    NUMPAD1: str = "{VK_NUMPAD1}"
    NUMPAD2: str = "{VK_NUMPAD2}"
    NUMPAD3: str = "{VK_NUMPAD3}"
    NUMPAD4: str = "{VK_NUMPAD4}"
    NUMPAD5: str = "{VK_NUMPAD5}"
    NUMPAD6: str = "{VK_NUMPAD6}"
    NUMPAD7: str = "{VK_NUMPAD7}"
    NUMPAD8: str = "{VK_NUMPAD8}"
    NUMPAD9: str = "{VK_NUMPAD9}"


# From: https://github.com/python-xlib/python-xlib/blob/4e8bbf8fc4941e5da301a8b3db8d27e98de68666/Xlib/keysymdef/miscellany.py
@dataclass
class LinuxKeyboardCodes:
    SPACE: str = "space"
    TAB: str = "Tab"
    ENTER: str = "Return"
    BACKSPACE: str = "BackSpace"
    DELETE: str = "Delete"
    ESCAPE: str = "Escape"
    LEFT_ARROW: str = "Left"
    RIGHT_ARROW: str = "Right"
    UP_ARROW: str = "Up"
    DOWN_ARROW: str = "Down"
    HOME: str = "Home"
    END: str = "End"
    PAGE_UP: str = "Page_Up"
    PAGE_DOWN: str = "Page_Down"
    PRINT_SCREEN: str = "Print"
    SCROLLLOCK: str = "Scroll_Lock"
    CAPSLOCK: str = "Caps_Lock"
    NUMLOCK: str = "Num_Lock"
    F1: str = "F1"
    F2: str = "F2"
    F3: str = "F3"
    F4: str = "F4"
    F5: str = "F5"
    F6: str = "F6"
    F7: str = "F7"
    F8: str = "F8"
    F9: str = "F9"
    F10: str = "F10"
    F11: str = "F11"
    F12: str = "F12"
    F13: str = "F13"
    F14: str = "F14"
    F15: str = "F15"
    F16: str = "F16"
    F17: str = "F17"
    F18: str = "F18"
    F19: str = "F19"
    F20: str = "F20"
    F21: str = "F21"
    F22: str = "F22"
    F23: str = "F23"
    F24: str = "F24"
    INSERT: str = "Insert"
    CONTROL: str = "Control_L"  # No normal Control -> Alias
    SHIFT: str = "Shift_L"  # No normal Shift -> Alias
    ALT: str = "Alt_L"  # No normal Alt -> Alias
    CANCEL: str = "Cancel"
    CLEAR: str = "Clear"
    EXECUTE: str = "Execute"
    KANA: str = "Kana_Lock"
    KANJI: str = "Kanji"
    LCONTROL: str = "Control_L"
    LMENU: str = "Menu"
    LSHIFT: str = "Shift_L"
    NEXT: str = "Next"
    PAUSE: str = "Pause"
    PRIOR: str = "Prior"
    RCONTROL: str = "Control_R"
    RETURN: str = "Return"
    RSHIFT: str = "Shift_R"
    SELECT: str = "Select"
    MENU_RIGHT: str = "Menu"  # No right Menu -> Alias
    ADD: str = "KP_Add"
    SUBTRACT: str = "KP_Subtract"
    DECIMAL: str = "KP_Decimal"
    MULTIPLY: str = "KP_Multiply"
    DIVIDE: str = "KP_Divide"
    SEPARATOR: str = "KP_Separator"
    NUMPAD0: str = "KP_0"
    NUMPAD1: str = "KP_1"
    NUMPAD2: str = "KP_2"
    NUMPAD3: str = "KP_3"
    NUMPAD4: str = "KP_4"
    NUMPAD5: str = "KP_5"
    NUMPAD6: str = "KP_6"
    NUMPAD7: str = "KP_7"
    NUMPAD8: str = "KP_8"
    NUMPAD9: str = "KP_9"


KeyboardCodes = WinKeyboardCodes if is_windows else LinuxKeyboardCodes
__all__ = ["SyncInput", "AsyncInput", "KeyboardCodes", "WinKeyboardCodes", "LinuxKeyboardCodes", "is_windows"]



from typing import Optional


class WindowClosedException(Exception):
    def __init__(self, message: Optional[str] = None, pid: Optional[int] = None):
        if not message:
            if pid:
                message = f"Interaction not possible due to stale/closed window with pid {pid}"
            else:
                message = "Interaction not possible due to stale/closed window"

        super().__init__(message)



import asyncio
import os
import re
import subprocess
import time
from typing import Any, List, Literal, Tuple

from Xlib import X, display
from Xlib.error import BadWindow
from Xlib.ext.xtest import fake_input
from Xlib.XK import string_to_keysym
from Xlib.xobject.drawable import Window

from cdp_patches.input.exceptions import WindowClosedException

# Every Common Symbol on a QWERTY Keyboard, Source: https://github.com/python-xlib/python-xlib/blob/4e8bbf8fc4941e5da301a8b3db8d27e98de68666/Xlib/keysymdef/latin1.py
# Dict Source: https://github.com/svenlr/swift-map/blob/main/mainloop.py#L459
symbol_dict = {
    "@": "at",
    "`": "grave",
    "\t": "Tab",
    "|": "bar",
    "\n": "Return",
    "\r": "Return",
    "~": "asciitilde",
    "{": "braceleft",
    "[": "bracketleft",
    "]": "bracketright",
    "\\": "backslash",
    "_": "underscore",
    "^": "asciicircum",
    "!": "exclam",
    " ": "space",
    "#": "numbersign",
    '"': "quotedbl",
    "%": "percent",
    "$": "dollar",
    "'": "apostrophe",
    "&": "ampersand",
    ")": "parenright",
    "(": "parenleft",
    "+": "plus",
    "*": "asterisk",
    "-": "minus",
    ",": "comma",
    "/": "slash",
    ".": "period",
    "\\e": "Escape",
    "}": "braceright",
    ";": "semicolon",
    ":": "colon",
    "=": "equal",
    "<": "less",
    "?": "question",
    ">": "greater",
}

layouts_shifted_chars = {
    "QWERTY": '~!@#$%^&*()_+{}|:"<>?',
    "AZERTY": '^¨£$¤%µ§*()_+{}|:"<>?',
    "QWERTZ": '°!"§$%&/()=?`´*_:;<>|',
    "DVORAK": '~!@#$%^&*()_+{}|:"<>?',
    "COLEMAK": '~!@#$%^&*()_+{}|:"<>?',
    "WORKMAN": '~!@#$%^&*()_+{}|:"<>?',
    "BEPO": '~!@#$%^&*()_+{}|:"<>?',
    "NEO": '~!@#$%^&*()_+{}|:"<>?',
    "JCUKEN": '!"№;%:?*()_+{}|:"<>?',
    "RUSSIAN": '!"№;%:?*()_+{}|:"<>?',
}


# Note for someday: `setxkbmap us`
def get_kb_layout_shifted_chars() -> str:
    search_regex = r'xkb_keycodes\s+{\s+include\s+"[A-Za-z\+]*\((.*?)\)"'
    layout_info = subprocess.check_output(["setxkbmap", "-print"]).decode("utf-8")
    matches = re.search(search_regex, layout_info)
    if matches:
        layout = matches.group(1).upper()
        if layout in layouts_shifted_chars:
            return layouts_shifted_chars[layout]
        raise EnvironmentError(f"Keyboard layout not recognized: {layout}!")
    raise EnvironmentError("Keyboard layout not found!")


class LinuxBase:
    browser_window: Window
    hwnd: int
    pid: int
    scale_factor: float = 1.0
    toolbar_height: int = 0
    render_widget_height: int = 0
    shifted_chars = get_kb_layout_shifted_chars()

    def __init__(self, pid: int, scale_factor: float) -> None:
        self.pid = pid
        self.scale_factor = scale_factor
        self._loop = asyncio.get_event_loop()

        display_env = os.getenv("DISPLAY")
        self.display = display.Display(display_env)
        self.tab_pid = self.get_window()

        self.browser_window = self.display.create_resource_object("window", self.tab_pid)

    def get_window(self) -> Any:
        name_atom = self.display.get_atom("WM_NAME", only_if_exists=True)
        pid_atom = self.display.get_atom("_NET_WM_PID", only_if_exists=True)
        res_windows: List[Window] = []

        # Getting all WindowIds by PID by recursively searching through all windows under the root window query tree
        def search_windows_by_pid(query_tree, pid: int):
            for window in query_tree.children:
                window_pid = window.get_property(pid_atom, 0, 0, pow(2, 32) - 1)
                if window_pid and window_pid.value[0] == pid:
                    res_windows.append(window)
                if window.query_tree().children:
                    search_windows_by_pid(window.query_tree(), pid)

        search_windows_by_pid(self.display.screen().root.query_tree(), self.pid)
        if not res_windows:
            raise WindowClosedException(f"No windows found for PID: {self.pid}")

        for window in res_windows:
            # Getting necessary window properties
            title = window.get_property(name_atom, 0, 0, pow(2, 32) - 1).value
            min_height = window.get_wm_normal_hints().min_height
            # parent_offset_coords = window.translate_coords(window.query_tree().parent, 0, 0)
            # window_x, window_y = parent_offset_coords.x, parent_offset_coords.y

            # Filter out non-browser windows, for example the Taskbar or Info Bars
            if (b"google-chrome" in title) or (title == b"chrome") or (min_height == 0):  # or (window_x == window_y) or not all((window_x, window_y))
                continue

            self.browser_window = window
            return self.browser_window

        raise WindowClosedException(f"No windows found for PID: {self.pid}")

    def ensure_window(self) -> None:
        try:
            # No Easy Visibility Check, well just check a random window attribute...
            if not self.browser_window.get_wm_normal_hints().min_height:
                self.get_window()
        except BadWindow:
            self.get_window()

    async def async_get_window(self) -> Any:
        name_atom = self.display.get_atom("WM_NAME", only_if_exists=True)
        pid_atom = self.display.get_atom("_NET_WM_PID", only_if_exists=True)
        res_windows: List[Window] = []

        # Getting all WindowIds by PID by recursively searching through all windows under the root window query tree
        def search_windows_by_pid(query_tree, pid: int):
            for window in query_tree.children:
                window_pid = window.get_property(pid_atom, 0, 0, pow(2, 32) - 1)
                if window_pid and window_pid.value[0] == pid:
                    res_windows.append(window)
                if window.query_tree().children:
                    search_windows_by_pid(window.query_tree(), pid)

        await self._loop.run_in_executor(None, lambda: search_windows_by_pid(self.display.screen().root.query_tree(), self.pid))
        if not res_windows:
            raise WindowClosedException(f"No windows found for PID: {self.pid}")

        for window in res_windows:
            # Getting necessary window properties
            title = window.get_property(name_atom, 0, 0, pow(2, 32) - 1).value
            min_height = window.get_wm_normal_hints().min_height
            # parent_offset_coords = window.translate_coords(window.query_tree().parent, 0, 0)
            # window_x, window_y = parent_offset_coords.x, parent_offset_coords.y

            # Filter out non-browser windows, for example the Taskbar or Info Bars
            if (b"google-chrome" in title) or (title == b"chrome") or (min_height == 0):  # or (window_x == window_y) or not all((window_x, window_y))
                continue

            self.browser_window = window
            return self.browser_window

        raise WindowClosedException(f"No windows found for PID: {self.pid}")

    def _offset_toolbar_height(self) -> Tuple[int, int]:
        # Get Window Location
        root_offset_coords = self.browser_window.translate_coords(self.browser_window.query_tree().root, 0, 0)
        parent_offset_coords = self.browser_window.translate_coords(self.browser_window.query_tree().parent, 0, 0)
        window_x = abs(root_offset_coords.x) + abs(parent_offset_coords.x)
        window_y = abs(root_offset_coords.y) + abs(parent_offset_coords.y)

        # Get Chrome Toolbar Height (Note: Fetching Chromes "Program Specified Minimum Size" - 1 for Chrome Toolbar Height + 1 for the Minimum Tab Size)
        chrome_toolbar_height = self.browser_window.get_wm_normal_hints().min_height - 1

        # Get Linux (Outer) Window Toolbar Height
        frame_extends_atom = self.display.get_atom("_NET_FRAME_EXTENTS", only_if_exists=True)
        if frame_extends_atom == X.NONE:
            raise ValueError('No Atom interned with the Name "_NET_FRAME_EXTENTS".')

        net_frame_extends = self.browser_window.get_property(frame_extends_atom, 0, 0, pow(2, 32) - 1)
        if net_frame_extends:
            window_toolbar_height = net_frame_extends.value[2]
            window_toolbar_width = net_frame_extends.value[3]
        else:
            window_toolbar_height = 0
            window_toolbar_width = 0

        offset_width: int = window_x - window_toolbar_width
        offset_height: int = window_y - window_toolbar_height + chrome_toolbar_height
        return offset_width, offset_height

    @staticmethod
    def _translate_button(button: Literal["left", "right", "middle", "scroll_up", "scroll_down"]) -> int:
        if button == "left":
            return 1
        elif button == "middle":
            return 2
        elif button == "right":
            return 3
        elif button == "scroll_up":
            return 4
        elif button == "scroll_down":
            return 5

    def down(self, button: Literal["left", "right", "middle"], x: int, y: int) -> None:
        self.ensure_window()
        self.move(x=x, y=y)
        fake_input(self.display, X.ButtonPress, self._translate_button(button))
        self.display.sync()

    def up(self, button: Literal["left", "right", "middle"], x: int, y: int) -> None:
        self.ensure_window()
        self.move(x=x, y=y)
        fake_input(self.display, X.ButtonRelease, self._translate_button(button))
        self.display.sync()

    def move(self, x: int, y: int) -> None:
        self.ensure_window()
        offset_width, offset_height = self._offset_toolbar_height()
        x = int(x * self.scale_factor) + offset_width
        y = int(y * self.scale_factor) + offset_height

        fake_input(self.display, X.MotionNotify, x=x, y=y)
        self.display.sync()

    def scroll(self, direction: Literal["up", "down", "left", "right"], amount: int) -> None:
        self.ensure_window()
        if direction in ("left", "right"):
            raise NotImplementedError("Scrolling horizontally is not supported on Linux.")

        scroll_direction: Literal["scroll_up", "scroll_down"] = "scroll_up" if direction == "up" else "scroll_down"

        for _ in range(amount):
            fake_input(self.display, X.ButtonPress, self._translate_button(scroll_direction))
            self.display.sync()
            fake_input(self.display, X.ButtonRelease, self._translate_button(scroll_direction))
            self.display.sync()

    def send_keystrokes(self, text: str) -> None:
        self.ensure_window()
        selective_regex = re.compile(r"{[^{}]*}|.")  # Only for redundancy of windows implementations
        shift_keycode = self.display.keysym_to_keycode(0xFFE1)  # Shift Key (0xFFE1)

        for key in selective_regex.findall(text):
            shifted_key = key.isupper() or key in self.shifted_chars
            if key in symbol_dict:
                key = symbol_dict[key]

            keysym = string_to_keysym(key)
            keycode = self.display.keysym_to_keycode(keysym)
            self.browser_window.set_input_focus(X.RevertToNone, X.CurrentTime)

            if shifted_key:
                fake_input(self.display, X.KeyPress, shift_keycode)
                # time.sleep(0.1)

            fake_input(self.display, X.KeyPress, keycode)
            self.display.sync()
            time.sleep(0.01)  # Note: Might want to increase this in the future, to make it more human-like, but pywinauto uses the same timeouts so for now its fine.

            fake_input(self.display, X.KeyRelease, keycode)

            if shifted_key:
                # time.sleep(0.1)
                fake_input(self.display, X.KeyRelease, shift_keycode)
            self.display.sync()



from __future__ import annotations

import asyncio
import random
import re
import sys
import time
import warnings
from typing import Any, Generator, Literal, Optional, Union

if sys.version_info.minor >= 10:
    from typing import TypeAlias
else:
    TypeAlias = "TypeAlias"  # type: ignore[assignment]

from cdp_patches import is_windows
from cdp_patches.input.exceptions import WindowClosedException

if is_windows:
    from pywinauto.application import ProcessNotFoundError
    from pywinauto.base_wrapper import ElementNotEnabled

    from cdp_patches.input.os_base.windows import WindowsBase  # type: ignore[assignment]

    LinuxBase: TypeAlias = WindowsBase  # type: ignore[no-redef]
    InputBase = WindowsBase  # type: ignore
    WindowErrors = (ValueError, ElementNotEnabled, ProcessNotFoundError, WindowClosedException)  # type: ignore[assignment]
else:
    from cdp_patches.input.os_base.linux import LinuxBase  # type: ignore[assignment]

    WindowsBase: TypeAlias = LinuxBase  # type: ignore[no-redef]
    InputBase = LinuxBase  # type: ignore
    WindowErrors = (AssertionError, ValueError, WindowClosedException)  # type: ignore[assignment]

from .browsers import DriverlessAsyncChrome, async_browsers, get_async_browser_pid, get_async_scale_factor
from .mouse_trajectory import HumanizeMouseTrajectory


class AsyncInput:
    emulate_behaviour: Optional[bool] = True
    pid: Optional[int]
    _base: Union[WindowsBase, LinuxBase]
    window_timeout: float = 30.0
    _scale_factor: float = 1.0
    sleep_timeout: float = 0.01
    typing_speed: int = 50
    last_x: int = 0
    last_y: int = 0
    selective_modifiers_regex = re.compile(r"{[^{}]*}|.")

    def __init__(
        self, pid: Optional[int] = None, browser: Optional[async_browsers] = None, scale_factor: Optional[float] = 1.0, emulate_behaviour: Optional[bool] = True, window_timeout: Optional[float] = 30.0
    ) -> None:
        self.pid = pid
        self.browser = browser
        self.window_timeout = window_timeout or self.window_timeout
        self._scale_factor = scale_factor or self._scale_factor
        self.emulate_behaviour = emulate_behaviour or self.emulate_behaviour
        self._move_lock = asyncio.Lock()

    def __await__(self) -> Generator[None, Any, AsyncInput]:
        yield from self.__ainit__().__await__()
        return self

    async def __ainit__(self) -> None:
        if self.browser:
            self.pid = await get_async_browser_pid(self.browser)
            self._scale_factor = float(await get_async_scale_factor(self.browser))
        elif not self.pid:
            raise ValueError("You must provide a pid or a browser")

        self._base = InputBase(self.pid, self._scale_factor)  # type: ignore
        await self._wait_for_window()

        # Include Windows Scale Factor for every browser except DriverlessSyncChrome
        if is_windows and not isinstance(self.browser, DriverlessAsyncChrome):
            self._base.include_windows_scale_factor()

    @property
    def base(self) -> Union[WindowsBase, LinuxBase]:
        return self._base

    @property
    def scale_factor(self) -> float:
        return self._scale_factor

    @scale_factor.setter
    def scale_factor(self, scale_value: float) -> None:
        self._scale_factor = scale_value
        if self._base:
            self._base.scale_factor = scale_value

    async def _wait_for_window(self) -> None:
        max_wait = time.perf_counter() + self.window_timeout
        while time.perf_counter() < max_wait:
            try:
                if await self.base.async_get_window():
                    return
            except WindowErrors:
                pass
            await self._sleep_timeout(0.1)

        raise TimeoutError(f"Chrome Window (PID: {self.pid}) not found in {self.window_timeout} seconds.")

    async def _sleep_timeout(self, timeout: Optional[float] = None) -> None:
        timeout = timeout or self.sleep_timeout
        if not random.randint(0, 10):
            timeout_random = self.sleep_timeout / 10
            timeout = timeout or random.uniform(self.sleep_timeout, self.sleep_timeout + timeout_random)

        await asyncio.sleep(timeout)

    async def click(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None) -> None:
        x, y = int(x), int(y)

        await self.down(button=button, x=x, y=y, emulate_behaviour=emulate_behaviour, timeout=timeout)
        if self.emulate_behaviour and emulate_behaviour:
            await self._sleep_timeout(timeout=timeout)
        await self.up(button=button, x=x, y=y)
        self.last_x, self.last_y = x, y

    async def double_click(
        self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None
    ) -> None:
        x, y = int(x), int(y)

        await self.click(button=button, x=x, y=y, timeout=timeout, emulate_behaviour=emulate_behaviour)
        if self.emulate_behaviour and emulate_behaviour:
            await self._sleep_timeout(random.uniform(0.14, 0.21))
            # await self._sleep_timeout(timeout=timeout)
        await self.click(button=button, x=x, y=y, emulate_behaviour=False, timeout=timeout)

        self.last_x, self.last_y = x, y

    async def down(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None) -> None:
        x, y = int(x), int(y)

        if self.emulate_behaviour and emulate_behaviour:
            await self.move(x=x, y=y, timeout=timeout, emulate_behaviour=emulate_behaviour)
        self._base.down(button=button, x=x, y=y)
        self.last_x, self.last_y = x, y

    async def up(self, button: Literal["left", "right", "middle"], x: Union[int, float], y: Union[int, float]) -> None:
        x, y = int(x), int(y)

        self._base.up(button=button, x=x, y=y)
        self.last_x, self.last_y = x, y

    async def move(self, x: Union[int, float], y: Union[int, float], emulate_behaviour: Optional[bool] = True, timeout: Optional[float] = None) -> None:
        async with self._move_lock:
            x, y = int(x), int(y)

            if self.emulate_behaviour and emulate_behaviour:
                humanized_points = HumanizeMouseTrajectory((self.last_x, self.last_y), (x, y))

                # Move Mouse to new random locations
                for i, (human_x, human_y) in enumerate(humanized_points.points):
                    self._base.move(x=int(human_x), y=int(human_y))
                    await self._sleep_timeout(timeout=timeout)

            else:
                self._base.move(x=x, y=y)
            self.last_x, self.last_y = x, y

    async def scroll(self, direction: Literal["up", "down", "left", "right"], amount: int) -> None:
        warnings.warn("Scrolling using CDP-Patches is discouraged as Scroll Inputs dont leak the CDP Domain.", UserWarning)
        self._base.scroll(direction=direction, amount=amount)

    async def type(self, text: str, fill: Optional[bool] = False, timeout: Optional[float] = None) -> None:
        if self.emulate_behaviour and not fill:
            for i, char in enumerate(self.selective_modifiers_regex.findall(text)):
                # If new word is started wait some more time
                if i != 0 and text[i - 1] == " ":
                    await self._sleep_timeout(timeout=timeout)

                self._base.send_keystrokes(char)
                await self._sleep_timeout((random.random() * 10) / self.typing_speed)
        else:
            self._base.send_keystrokes(text)
